"""
minfraud
~~~~~~~~

A client API to MaxMind's minFraud Score and Insights web services.
"""

from .errors import MinFraudError, AuthenticationError, \
    HTTPError, InvalidRequestError, InsufficientFundsError

from .webservice import Client
from .version import __version__

__author__ = 'Gregory Oschwald'



"""
minfraud.webservice
~~~~~~~~~~~~~~~~~~~

This module contains the webservice client class.

"""

import requests
from requests.utils import default_user_agent
from voluptuous import MultipleInvalid

from .version import __version__
from .errors import MinFraudError, HTTPError, AuthenticationError, \
    InsufficientFundsError, InvalidRequestError, PermissionRequiredError
from .models import Factors, Insights, Score
from .validation import validate_transaction


class Client(object):
    """Client for accessing the minFraud Score and Insights web services."""
    def __init__(
            self,
            account_id=None,
            license_key=None,
            host='minfraud.maxmind.com',
            locales=('en', ),
            timeout=None,

            # This is deprecated and not documented for that reason.
            # It can be removed if we do a major release in the future.
            user_id=None):
        """Constructor for Client.

        :param account_id: Your MaxMind account ID
        :type account_id: int
        :param license_key: Your MaxMind license key
        :type license_key: str
        :param host: The host to use when connecting to the web service.
        :type host: str
        :param locales: A tuple of locale codes to use in name property
        :type locales: tuple[str]
        :param timeout: The timeout to use for the request.
        :type timeout: float
        :return: Client object
        :rtype: Client
        """
        # pylint: disable=too-many-arguments
        if account_id is None:
            account_id = user_id

        if account_id is None:
            raise TypeError('The account_id is a required parameter')
        if license_key is None:
            raise TypeError('The license_key is a required parameter')

        # pylint: disable=too-many-arguments
        self._locales = locales
        # requests 2.12.2 requires that the username passed to auth be a
        # string
        self._account_id = account_id if isinstance(account_id,
                                                    bytes) else str(account_id)
        self._license_key = license_key
        self._base_uri = u'https://{0:s}/minfraud/v2.0'.format(host)
        self._timeout = timeout

    def factors(self, transaction, validate=True):
        """Query Factors endpoint with transaction data.

        :param transaction: A dictionary containing the transaction to be
          sent to the minFraud Insights web service as specified in the `REST
          API documentation
          <https://dev.maxmind.com/minfraud/#Request_Body>`_.
        :type transaction: dict
        :param validate: If set to false, validation of the transaction
          dictionary will be disabled. This validation helps ensure that your
          request is correct before sending it to MaxMind. Validation raises an
          InvalidRequestError.
        :type validate: bool
        :return: A Factors model object
        :rtype: Factors
        :raises: AuthenticationError, InsufficientFundsError,
          InvalidRequestError, HTTPError, MinFraudError,
        """
        return self._response_for('factors', Factors, transaction, validate)

    def insights(self, transaction, validate=True):
        """Query Insights endpoint with transaction data.

        :param transaction: A dictionary containing the transaction to be
          sent to the minFraud Insights web service as specified in the `REST
          API documentation
          <https://dev.maxmind.com/minfraud/#Request_Body>`_.
        :type transaction: dict
        :param validate: If set to false, validation of the transaction
          dictionary will be disabled. This validation helps ensure that your
          request is correct before sending it to MaxMind. Validation raises an
          InvalidRequestError.
        :type validate: bool
        :return: An Insights model object
        :rtype: Insights
        :raises: AuthenticationError, InsufficientFundsError,
          InvalidRequestError, HTTPError, MinFraudError,
        """
        return self._response_for('insights', Insights, transaction, validate)

    def score(self, transaction, validate=True):
        """Query Score endpoint with transaction data.

        :param transaction: A dictionary containing the transaction to be
          sent to the minFraud Score web service as specified in the `REST API
          documentation
          <https://dev.maxmind.com/minfraud/#Request_Body>`_.
        :type transaction: dict
        :param validate: If set to false, validation of the transaction
          dictionary will be disabled. This validation helps ensure that your
          request is correct before sending it to MaxMind. Validation raises an
          InvalidRequestError.
        :type validate: bool
        :return: A Score model object
        :rtype: Score
        :raises: AuthenticationError, InsufficientFundsError,
          InvalidRequestError, HTTPError, MinFraudError,
        """
        return self._response_for('score', Score, transaction, validate)

    def _response_for(self, path, model_class, request, validate):
        """Send request and create response object."""
        cleaned_request = self._copy_and_clean(request)
        if validate:
            try:
                validate_transaction(cleaned_request)
            except MultipleInvalid as ex:
                raise InvalidRequestError(
                    "Invalid transaction data: {0}".format(ex))
        uri = '/'.join([self._base_uri, path])
        response = requests.post(uri,
                                 json=cleaned_request,
                                 auth=(self._account_id, self._license_key),
                                 headers={
                                     'Accept': 'application/json',
                                     'User-Agent': self._user_agent()
                                 },
                                 timeout=self._timeout)
        if response.status_code != 200:
            raise self._exception_for_error(response, uri)
        return self._handle_success(response, uri, model_class)

    def _copy_and_clean(self, data):
        """Create a copy of the data structure with Nones removed."""
        if isinstance(data, dict):
            return dict((k, self._copy_and_clean(v))
                        for (k, v) in data.items() if v is not None)
        if isinstance(data, (list, set, tuple)):
            return [self._copy_and_clean(x) for x in data if x is not None]
        return data

    def _user_agent(self):
        """Create User-Agent header."""
        return 'minFraud-API/%s %s' % (__version__, default_user_agent())

    def _handle_success(self, response, uri, model_class):
        """Handle successful response."""
        try:
            body = response.json()
        except ValueError:
            raise MinFraudError(
                'Received a 200 response'
                ' but could not decode the response as '
                'JSON: {0}'.format(response.content), 200, uri)
        if 'ip_address' in body:
            body['ip_address']['_locales'] = self._locales
        return model_class(body)

    def _exception_for_error(self, response, uri):
        """Returns the exception for the error responses."""
        status = response.status_code

        if 400 <= status < 500:
            return self._exception_for_4xx_status(response, status, uri)
        if 500 <= status < 600:
            return self._exception_for_5xx_status(status, uri)
        return self._exception_for_non_200_status(status, uri)

    def _exception_for_4xx_status(self, response, status, uri):
        """Returns exception for error responses with 4xx status codes."""
        if not response.content:
            return HTTPError(
                'Received a {0} error with no body'.format(status), status,
                uri)
        if response.headers.get('Content-Type', '').find('json') == -1:
            return HTTPError(
                'Received a {0} with the following '
                'body: {1}'.format(status, response.content), status, uri)
        try:
            body = response.json()
        except ValueError:
            return HTTPError(
                'Received a {status:d} error but it did not include'
                ' the expected JSON body: {content}'.format(
                    status=status, content=response.content), status, uri)
        else:
            if 'code' in body and 'error' in body:
                return self._exception_for_web_service_error(
                    body.get('error'), body.get('code'), status, uri)
            return HTTPError(
                'Error response contains JSON but it does not specify code'
                ' or error keys: {0}'.format(response.content), status, uri)

    def _exception_for_web_service_error(self, message, code, status, uri):
        """Returns exception for error responses with the JSON body."""
        if code in ('ACCOUNT_ID_REQUIRED', 'AUTHORIZATION_INVALID',
                    'LICENSE_KEY_REQUIRED', 'USER_ID_REQUIRED'):
            return AuthenticationError(message)
        if code == 'INSUFFICIENT_FUNDS':
            return InsufficientFundsError(message)
        if code == 'PERMISSION_REQUIRED':
            return PermissionRequiredError(message)

        return InvalidRequestError(message, code, status, uri)

    def _exception_for_5xx_status(self, status, uri):
        """Returns exception for error response with 5xx status codes."""
        return HTTPError(
            u'Received a server error ({0}) for '
            u'{1}'.format(status, uri), status, uri)

    def _exception_for_non_200_status(self, status, uri):
        """Returns exception for responses with unexpected status codes."""
        return HTTPError(
            u'Received an unexpected HTTP status '
            u'({0}) for {1}'.format(status, uri), status, uri)



from decimal import Decimal
import sys
from voluptuous import MultipleInvalid

from minfraud.validation import validate_transaction

if sys.version_info[:2] == (2, 6):
    import unittest2 as unittest
else:
    import unittest

if sys.version_info[0] == 2:
    unittest.TestCase.assertRaisesRegex = unittest.TestCase.assertRaisesRegexp
    unittest.TestCase.assertRegex = unittest.TestCase.assertRegexpMatches


class ValidationBase(object):
    def setup_transaction(self, transaction):
        if 'device' not in transaction:
            transaction['device'] = {}

        if 'ip_address' not in transaction['device']:
            transaction['device']['ip_address'] = '1.1.1.1'

    def check_invalid_transaction(self, transaction):
        self.setup_transaction(transaction)
        with self.assertRaises(MultipleInvalid,
                               msg='{0!s} is invalid'.format(transaction)):
            validate_transaction(transaction)

    def check_transaction(self, transaction):
        self.setup_transaction(transaction)
        try:
            validate_transaction(transaction)
        except MultipleInvalid as e:
            self.fail('MultipleInvalid {0} thrown for {1}'.format(
                e.msg, transaction))

    def check_str_type(self, object, key):
        self.check_transaction({object: {key: 'string'}})
        self.check_invalid_transaction({object: {key: 12}})

    def check_positive_number(self, f):
        for good in (1, 1.1, Decimal('1.1')):
            self.check_transaction(f(good))
        for bad in ('1.2', '1', -1, -1.1, 0):
            self.check_invalid_transaction(f(bad))

    def check_bool(self, object, key):
        for good in (True, False):
            self.check_transaction({object: {key: good}})
        for bad in ('', 0, 'True'):
            self.check_invalid_transaction({object: {key: bad}})


class TestAccount(unittest.TestCase, ValidationBase):
    def test_account_user_id(self):
        self.check_transaction({'account': {'user_id': 'usr'}})

    def test_account_username_md5(self):
        self.check_transaction(
            {'account': {
                'username_md5': '14c4b06b824ec593239362517f538b29'
            }})

    def test_invalid_account_username_md5s(self):
        self.check_invalid_transaction(
            {'account': {
                'username_md5': '14c4b06b824ec593239362517f538b2'
            }})
        self.check_invalid_transaction(
            {'account': {
                'username_md5': '14c4b06b824ec593239362517f538b29a'
            }})


class AddressBase(ValidationBase):
    def test_strings(self):
        for key in ('first_name', 'last_name', 'company', 'address',
                    'address_2', 'city', 'postal', 'phone_number'):
            self.check_str_type(self.type, key)

    def test_region(self):
        for region in ('A', 'AA', 'AAA', 'ZZZZ'):
            self.check_transaction({self.type: {'region': region}})
        for invalid in ('', 'AAAAA', 1, 'aaa'):
            self.check_invalid_transaction({self.type: {'region': invalid}})

    def test_country(self):
        for country in ('US', 'CA', 'GB'):
            self.check_transaction({self.type: {'country': country}})
        for invalid in ('', 'U1', 'USA', 1, '11', 'us'):
            self.check_invalid_transaction({self.type: {'country': invalid}})

    def test_phone_country_code(self):
        for code in (1, '1', '2341'):
            self.check_transaction({self.type: {'phone_country_code': code}})
        for invalid in ('', '12345', 'U'):
            self.check_invalid_transaction(
                {self.type: {
                    'phone_country_code': invalid
                }})


class TestBillingAddress(unittest.TestCase, AddressBase):
    type = 'billing'


class TestShippingAddress(unittest.TestCase, AddressBase):
    type = 'shipping'

    def test_delivery_speed(self):
        for speed in ('same_day', 'overnight', 'expedited', 'standard'):
            self.check_transaction({self.type: {'delivery_speed': speed}})
        for invalid in ('fast', 'slow', ''):
            self.check_invalid_transaction(
                {self.type: {
                    'delivery_speed': invalid
                }})


class TestCreditCard(ValidationBase, unittest.TestCase):
    def test_issuer_id_number(self):
        for iin in ('123456', '532313'):
            self.check_transaction({'credit_card': {'issuer_id_number': iin}})
        for invalid in ('12345', '1234567', 123456, '12345a'):
            self.check_invalid_transaction(
                {'credit_card': {
                    'issuer_id_number': invalid
                }})

    def test_last_4_digits(self):
        for iin in ('1234', '9323'):
            self.check_transaction({'credit_card': {'last_4_digits': iin}})
        for invalid in ('12345', '123', 1234, '123a'):
            self.check_invalid_transaction(
                {'credit_card': {
                    'last_4_digits': invalid
                }})

    def test_bank_name(self):
        self.check_str_type('credit_card', 'bank_name')

    def test_bank_phone_number(self):
        self.check_str_type('credit_card', 'bank_phone_number')

    def test_phone_country_code(self):
        for code in (1, '1', '2341'):
            self.check_transaction(
                {'credit_card': {
                    'bank_phone_country_code': code
                }})
        for invalid in ('', '12345', 'U'):
            self.check_invalid_transaction(
                {'credit_card': {
                    'bank_phone_country_code': invalid
                }})

    def test_avs_and_cvv(self):
        for key in ('avs_result', 'cvv_result'):
            for code in ('1', 'A'):
                self.check_transaction({'credit_card': {key: code}})
            for invalid in ('', '12'):
                self.check_invalid_transaction(
                    {'credit_card': {
                        'credit_card': invalid
                    }})

    def test_token(self):
        for token in ('123456abc1245', '\x21', '1' * 20):
            self.check_transaction({'credit_card': {'token': token}})
        for invalid in ('\x20', '123456', 'x' * 256):
            self.check_invalid_transaction({'credit_card': {'token': invalid}})


class TestCustomInputs(ValidationBase, unittest.TestCase):
    def test_valid_inputs(self):
        self.check_transaction({
            'custom_inputs': {
                'string_input_1': 'test string',
                'int_input': 19,
                'float_input': 3.2,
                'bool_input': True
            }
        })

    def test_invalid(self):
        for invalid in (
            {'InvalidKey': 1},
            {'too_long': 'x' * 256},
            {'has_newline': 'test\n'},
            {'too_big': 1e13},
            {'too_small': -1e13},
            {'too_big_float': float(1e13)}
            ): # yapf: disable
            self.check_invalid_transaction({'custom_inputs': invalid})


class TestDevice(ValidationBase, unittest.TestCase):
    def test_ip_address(self):
        for ip in ('1.2.3.4', '2001:db8:0:0:1:0:0:1', '::FFFF:1.2.3.4'):
            self.check_transaction({'device': {'ip_address': ip}})
        for invalid in ('1.2.3.', '299.1.1.1', '::AF123'):
            self.check_invalid_transaction({'device': {'ip_address': invalid}})

    def test_missing_ip(self):
        with self.assertRaises(MultipleInvalid):
            validate_transaction({'device': {}})

    def test_missing_device(self):
        with self.assertRaises(MultipleInvalid):
            validate_transaction({})

    def test_user_agent(self):
        self.check_str_type('device', 'user_agent')

    def test_accept_language(self):
        self.check_str_type('device', 'accept_language')

    def test_session_id(self):
        self.check_str_type('device', 'session_id')

    def test_session_age(self):
        for valid in (3600, 0, 25.5):
            self.check_transaction(
                {'device': {
                    'ip_address': '4.4.4.4',
                    'session_age': valid
                }})
        for invalid in ('foo', -1):
            self.check_invalid_transaction(
                {'device': {
                    'ip_address': '4.4.4.4',
                    'session_age': invalid
                }})


class TestEmail(ValidationBase, unittest.TestCase):
    def test_address(self):
        for good in ('test@maxmind.com', '977577b140bfb7c516e4746204fbdb01'):
            self.check_transaction({'email': {'address': good}})
        for bad in ('not.email', '977577b140bfb7c516e4746204fbdb0',
                    '977577b140bfb7c516e4746204fbdb012'):
            self.check_invalid_transaction({'email': {'address': bad}})

    def test_domain(self):
        for good in ('maxmind.com', 'www.bbc.co.uk'):
            self.check_transaction({'email': {'domain': good}})
        for bad in ('bad ', ' bad.com'):
            self.check_invalid_transaction({'email': {'domain': bad}})


class TestEvent(ValidationBase, unittest.TestCase):
    def test_transaction(self):
        self.check_str_type('event', 'transaction_id')

    def test_shop_id(self):
        self.check_str_type('event', 'shop_id')

    def test_time(self):
        for good in ('2015-05-08T16:07:56+00:00', '2015-05-08T16:07:56Z'):
            self.check_transaction({'event': {'time': good}})
        for bad in ('2015-05-08T16:07:56', '2015-05-08 16:07:56Z'):
            self.check_invalid_transaction({'event': {'time': bad}})

    def test_type(self):
        for good in ('account_creation', 'account_login', 'email_change',
                     'password_reset', 'payout_change', 'purchase',
                     'recurring_purchase', 'referral', 'survey'):
            self.check_transaction({'event': {'type': good}})
        for bad in ('bad', 1, ''):
            self.check_invalid_transaction({'event': {'type': bad}})


class TestOrder(ValidationBase, unittest.TestCase):
    def test_amount(self):
        self.check_positive_number(lambda v: {'order': {'amount': v}})

    def test_currency(self):
        for good in ('USD', 'GBP'):
            self.check_transaction({'order': {'currency': good}})
        for bad in ('US', 'US1', 'USDD', 'usd'):
            self.check_invalid_transaction({'order': {'currency': bad}})

    def test_discount_code(self):
        self.check_str_type('order', 'discount_code')

    def test_affiliate_id(self):
        self.check_str_type('order', 'affiliate_id')

    def test_subaffiliate_id(self):
        self.check_str_type('order', 'subaffiliate_id')

    def test_is_gift(self):
        self.check_bool('order', 'is_gift')

    def test_has_gift_message(self):
        self.check_bool('order', 'has_gift_message')

    def test_referrer_uri(self):
        for good in ('http://www.mm.com/fadsf', 'https://x.org/'):
            self.check_transaction({'order': {'referrer_uri': good}})
        for bad in ('ftp://a.com/', 'www.mm.com'):
            self.check_invalid_transaction({'order': {'referrer_uri': bad}})


class TestPayment(ValidationBase, unittest.TestCase):
    def test_processor(self):
        for good in ('adyen', 'stripe'):
            self.check_transaction({'payment': {'processor': good}})
        for bad in ('notvalid', ' stripe'):
            self.check_invalid_transaction({'payment': {'processor': bad}})

    def test_was_authorized(self):
        self.check_bool('payment', 'was_authorized')

    def test_decline_code(self):
        self.check_str_type('payment', 'decline_code')


class TestShoppingCart(ValidationBase, unittest.TestCase):
    def test_category(self):
        self.check_transaction({'shopping_cart': [{'category': 'cat'}]})

    def test_item_id(self):
        self.check_transaction({'shopping_cart': [{'item_id': 'cat'}]})

    def test_amount(self):
        self.check_positive_number(lambda v: {'shopping_cart': [{'price': v}]})

    def test_quantity(self):
        for good in (1, 1000):
            self.check_transaction({'shopping_cart': [{'quantity': good}]})
        for bad in (1.1, -1, 0):
            self.check_invalid_transaction(
                {'shopping_cart': [{
                    'quantity': bad
                }]})



import os
import sys

import json
import requests_mock
from io import open
from minfraud.errors import HTTPError, InvalidRequestError, \
    AuthenticationError, InsufficientFundsError, MinFraudError, \
    PermissionRequiredError
from minfraud.models import Factors, Insights, Score
from minfraud.webservice import Client

if sys.version_info[:2] == (2, 6):
    import unittest2 as unittest
else:
    import unittest

if sys.version_info[0] == 2:
    unittest.TestCase.assertRaisesRegex = unittest.TestCase.assertRaisesRegexp
    unittest.TestCase.assertRegex = unittest.TestCase.assertRegexpMatches


class BaseTest(object):
    def setUp(self):
        self.client = Client(42, 'abcdef123456')
        test_dir = os.path.join(os.path.dirname(__file__), 'data')
        with open(os.path.join(test_dir, 'full-request.json'),
                  encoding='utf-8') as file:
            content = file.read()
        self.full_request = json.loads(content)
        with open(os.path.join(test_dir,
                               '{0}-response.json'.format(self.type)),
                  encoding='utf-8') as file:
            self.response = file.read()

    base_uri = 'https://minfraud.maxmind.com/minfraud/v2.0/'

    def has_ip_location(self):
        return self.type in ['factors', 'insights']

    def test_200(self):
        model = self.create_success()
        response = json.loads(self.response)
        if self.has_ip_location():
            response['ip_address']['_locales'] = ('en', )
        self.assertEqual(self.cls(response), model)
        if self.has_ip_location():
            self.assertEqual('United Kingdom', model.ip_address.country.name)

    def test_200_on_request_with_nones(self):
        model = self.create_success(
            request={
                'device': {
                    'ip_address': '81.2.69.160',
                    'accept_language': None
                },
                'event': {
                    'shop_id': None
                },
                'shopping_cart': [{
                    'category': None,
                    'quantity': 2,
                }, None],
            })
        response = self.response
        self.assertEqual(0.01, model.risk_score)

    def test_200_with_locales(self):
        locales = ('fr', )
        client = Client(42, 'abcdef123456', locales=locales)
        model = self.create_success(client=client)
        response = json.loads(self.response)
        if self.has_ip_location():
            response['ip_address']['_locales'] = locales
        self.assertEqual(self.cls(response), model)
        if self.has_ip_location():
            self.assertEqual('Royaume-Uni', model.ip_address.country.name)
            self.assertEqual('Londres', model.ip_address.city.name)

    def test_200_with_no_body(self):
        with self.assertRaisesRegex(
                MinFraudError,
                "Received a 200 response but could not decode the response as"
                " JSON: b?'?'?"):
            self.create_success(text='')

    def test_200_with_invalid_json(self):
        with self.assertRaisesRegex(
                MinFraudError,
                "Received a 200 response but could not decode the response as"
                " JSON: b?'?{'?"):
            self.create_success(text='{')

    def test_insufficient_funds(self):
        with self.assertRaisesRegex(InsufficientFundsError, 'out of funds'):
            self.create_error(
                text='{"code":"INSUFFICIENT_FUNDS","error":"out of funds"}',
                status_code=402)

    def test_invalid_auth(self):
        for error in ('ACCOUNT_ID_REQUIRED', 'AUTHORIZATION_INVALID',
                      'LICENSE_KEY_REQUIRED', 'USER_ID_REQUIRED'):
            with self.assertRaisesRegex(AuthenticationError, 'Invalid auth'):
                self.create_error(
                    text=u'{{"code":"{0:s}","error":"Invalid auth"}}'.format(
                        error),
                    status_code=401)

    def test_invalid_request(self):
        with self.assertRaisesRegex(InvalidRequestError, 'IP invalid'):
            self.create_error(
                text='{"code":"IP_ADDRESS_INVALID","error":"IP invalid"}')

    def test_permission_required(self):
        with self.assertRaisesRegex(PermissionRequiredError, 'permission'):
            self.create_error(
                text=
                '{"code":"PERMISSION_REQUIRED","error":"permission required"}',
                status_code=403)

    def test_400_with_invalid_json(self):
        with self.assertRaisesRegex(
                HTTPError,
                "Received a 400 error but it did not include the expected JSON"
                " body: b?'?{blah}'?"):
            self.create_error(text='{blah}')

    def test_400_with_no_body(self):
        with self.assertRaisesRegex(HTTPError,
                                    'Received a 400 error with no body'):
            self.create_error()

    def test_400_with_unexpected_content_type(self):
        with self.assertRaisesRegex(
                HTTPError,
                "Received a 400 with the following body: b?'?plain'?"):
            self.create_error(headers={'Content-Type': 'text/plain'},
                              text='plain')

    def test_400_without_json_body(self):
        with self.assertRaisesRegex(
                HTTPError,
                "Received a 400 error but it did not include the expected JSON"
                " body: b?'?plain'?"):
            self.create_error(text='plain')

    def test_400_with_unexpected_json(self):
        with self.assertRaisesRegex(
                HTTPError,
                'Error response contains JSON but it does not specify code or'
                ' error keys: b?\'?{"not":"expected"}\'?'):
            self.create_error(text='{"not":"expected"}')

    def test_300_error(self):
        with self.assertRaisesRegex(
                HTTPError, 'Received an unexpected HTTP status \(300\) for'):
            self.create_error(status_code=300)

    def test_500_error(self):
        with self.assertRaisesRegex(HTTPError,
                                    'Received a server error \(500\) for'):
            self.create_error(status_code=500)

    @requests_mock.mock()
    def create_error(self, mock, status_code=400, text='', headers=None):
        if headers is None:
            headers = {
                'Content-Type':
                'application/vnd.maxmind.com-error+json; charset=UTF-8;'
                ' version=2.0'
            }
        mock.post(self.base_uri + self.type,
                  status_code=status_code,
                  text=text,
                  headers=headers)
        return getattr(self.client, self.type)(self.full_request)

    @requests_mock.mock()
    def create_success(self,
                       mock,
                       text=None,
                       headers=None,
                       client=None,
                       request=None):
        if headers is None:
            headers = {
                'Content-Type':
                'application/vnd.maxmind.com-minfraud-{0}+json;'
                ' charset=UTF-8; version=2.0'.format(self.type)
            }
        if text is None:
            text = self.response
        mock.post(self.base_uri + self.type,
                  status_code=200,
                  text=text,
                  headers=headers)
        if client is None:
            client = self.client
        if request is None:
            request = self.full_request
        return getattr(client, self.type)(request)

    def test_named_constructor_args(self):
        id = '47'
        key = '1234567890ab'
        for client in (Client(account_id=id, license_key=key),
                       Client(user_id=id, license_key=key)):
            self.assertEqual(client._account_id, id)
            self.assertEqual(client._license_key, key)

    def test_missing_constructor_args(self):
        with self.assertRaises(TypeError):
            Client(license_key='1234567890ab')

        with self.assertRaises(TypeError):
            Client('47')


class TestFactors(BaseTest, unittest.TestCase):
    type = 'factors'
    cls = Factors


class TestInsights(BaseTest, unittest.TestCase):
    type = 'insights'
    cls = Insights


class TestScore(BaseTest, unittest.TestCase):
    type = 'score'
    cls = Score



#!/usr/bin/env python

import ast
import io
import os
import re
import sys

from setuptools import setup

_version_re = re.compile(r'__version__\s+=\s+(.*)')

with io.open('minfraud/version.py', 'r', encoding='utf-8') as f:
    _version = str(ast.literal_eval(_version_re.search(f.read()).group(1)))

with io.open('README.rst', 'r', encoding='utf-8') as f:
    _readme = f.read()

requirements = [
    'geoip2>=3.0.0,<4.0.0',
    'requests>=2.22.0',
    'rfc3987',
    'strict-rfc3339',
    'urllib3>=1.25.2',
    'validate_email',
    'voluptuous',
]

# Write requirements.txt needed for snyk testing, only for latest release python.
if os.environ.get('SNYK_TOKEN') and os.environ.get('RUN_SNYK'):
    with open('requirements.txt', 'w') as f:
        for r in requirements:
            f.write(r + '\n')

setup(
    name='minfraud',
    version=_version,
    description='MaxMind minFraud Score, Insights, and Factors API',
    long_description=_readme,
    author='Gregory Oschwald',
    author_email='goschwald@maxmind.com',
    url='http://www.maxmind.com/',
    packages=['minfraud'],
    include_package_data=True,
    platforms='any',
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*',
    install_requires=requirements,
    extras_require={':python_version=="2.7"': ['ipaddress']},
    tests_require=['requests_mock'],
    test_suite="tests",
    license='Apache License 2.0 ',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python',
        'Topic :: Internet :: Proxy Servers',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet',
    ],
)



import sys

from minfraud.models import *

if sys.version_info[:2] == (2, 6):
    import unittest2 as unittest
else:
    import unittest

if sys.version_info[0] == 2:
    unittest.TestCase.assertRaisesRegex = unittest.TestCase.assertRaisesRegexp
    unittest.TestCase.assertRegex = unittest.TestCase.assertRegexpMatches


class TestModels(unittest.TestCase):
    def test_model_immutability(self):
        """This tests some level of _shallow_ immutability for these classes"""
        T = namedtuple('T', ['obj', 'attr'])
        models = [
            T(Issuer(), 'name'),
            T(CreditCard(), 'country'),
            T(Device(), 'id'),
            T(Email(), 'is_free'),
            T(BillingAddress(), 'latitude'),
            T(ShippingAddress(), 'latitude'),
            T(ServiceWarning(), 'code'),
            T(Insights(), 'id'),
            T(Score(), 'id'),
            T(IPAddress({}), 'city')
        ]
        for model in models:
            for attr in (model.attr, 'does_not_exist'):
                with self.assertRaises(AttributeError,
                                       msg='{0!s} - {0}'.format(
                                           model.obj, attr)):
                    setattr(model.obj, attr, 5)

    def test_billing_address(self):
        address = BillingAddress(self.address_dict)
        self.check_address(address)

    def test_shipping_address(self):
        address_dict = self.address_dict
        address_dict['is_high_risk'] = False
        address_dict['distance_to_billing_address'] = 200

        address = ShippingAddress(address_dict)
        self.check_address(address)
        self.assertEqual(False, address.is_high_risk)
        self.assertEqual(200, address.distance_to_billing_address)

    @property
    def address_dict(self):
        return {
            'is_in_ip_country': True,
            'latitude': 43.1,
            'longitude': 32.1,
            'distance_to_ip_location': 100,
            'is_postal_in_city': True
        }

    def check_address(self, address):
        self.assertEqual(True, address.is_in_ip_country)
        self.assertEqual(True, address.is_postal_in_city)
        self.assertEqual(100, address.distance_to_ip_location)
        self.assertEqual(32.1, address.longitude)
        self.assertEqual(43.1, address.latitude)

    def test_credit_card(self):
        cc = CreditCard({
            'issuer': {
                'name': 'Bank'
            },
            'brand': 'Visa',
            'country': 'US',
            'is_issued_in_billing_address_country': True,
            'is_prepaid': True,
            'is_virtual': True,
            'type': 'credit'
        })

        self.assertEqual('Bank', cc.issuer.name)
        self.assertEqual('Visa', cc.brand)
        self.assertEqual('US', cc.country)
        self.assertEqual(True, cc.is_prepaid)
        self.assertEqual(True, cc.is_virtual)
        self.assertEqual(True, cc.is_issued_in_billing_address_country)
        self.assertEqual('credit', cc.type)

    def test_device(self):
        id = 'b643d445-18b2-4b9d-bad4-c9c4366e402a'
        last_seen = '2016-06-08T14:16:38Z'
        local_time = '2016-06-10T14:19:10-08:00'
        device = Device({
            'confidence': 99,
            'id': id,
            'last_seen': last_seen,
            'local_time': local_time,
        })

        self.assertEqual(99, device.confidence)
        self.assertEqual(id, device.id)
        self.assertEqual(last_seen, device.last_seen)
        self.assertEqual(local_time, device.local_time)

    def test_disposition(self):
        disposition = Disposition({'action': 'accept', 'reason': 'default'})

        self.assertEqual('accept', disposition.action)
        self.assertEqual('default', disposition.reason)

    def test_email(self):
        first_seen = '2016-01-01'
        email = Email({
            'first_seen': first_seen,
            'is_free': True,
            'is_high_risk': False
        })

        self.assertEqual(first_seen, email.first_seen)
        self.assertEqual(True, email.is_free)
        self.assertEqual(False, email.is_high_risk)

    def test_geoip2_country(self):
        country = GeoIP2Country(is_high_risk=True, iso_code='US')
        self.assertEqual(True, country.is_high_risk)
        self.assertEqual('US', country.iso_code)

    def test_geoip2_location(self):
        time = "2015-04-19T12:59:23-01:00"
        location = GeoIP2Location(local_time=time, latitude=5)
        self.assertEqual(time, location.local_time)
        self.assertEqual(5, location.latitude)

    def test_ip_address(self):
        time = "2015-04-19T12:59:23-01:00"
        address = IPAddress({
            'country': {
                'is_high_risk': True,
                'is_in_european_union': True,
            },
            'location': {
                'local_time': time,
            },
            'risk': 99,
            'traits': {
                'is_anonymous': True,
                'is_anonymous_proxy': True,
                'is_anonymous_vpn': True,
                'is_hosting_provider': True,
                'is_public_proxy': True,
                'is_satellite_provider': True,
                'is_tor_exit_node': True,
            },
        })

        self.assertEqual(time, address.location.local_time)
        self.assertEqual(True, address.country.is_high_risk)
        self.assertEqual(True, address.country.is_in_european_union)
        self.assertEqual(99, address.risk)
        self.assertEqual(True, address.traits.is_anonymous)
        self.assertEqual(True, address.traits.is_anonymous_proxy)
        self.assertEqual(True, address.traits.is_anonymous_vpn)
        self.assertEqual(True, address.traits.is_hosting_provider)
        self.assertEqual(True, address.traits.is_public_proxy)
        self.assertEqual(True, address.traits.is_satellite_provider)
        self.assertEqual(True, address.traits.is_tor_exit_node)
        self.assertEqual(True, address.country.is_high_risk)

    def test_score_ip_address(self):
        address = ScoreIPAddress({'risk': 99})
        self.assertEqual(99, address.risk)

    def test_ip_address_locales(self):

        loc = IPAddress({
            '_locales': ['fr'],
            'country': {
                'names': {
                    'fr': 'Country'
                }
            },
            'city': {
                'names': {
                    'fr': 'City'
                }
            },
        })

        self.assertEqual('City', loc.city.name)
        self.assertEqual('Country', loc.country.name)

    def test_issuer(self):
        phone = '132-342-2131'

        issuer = Issuer({
            'name': 'Bank',
            'matches_provided_name': True,
            'phone_number': phone,
            'matches_provided_phone_number': True
        })

        self.assertEqual('Bank', issuer.name)
        self.assertEqual(True, issuer.matches_provided_name)
        self.assertEqual(phone, issuer.phone_number)
        self.assertEqual(True, issuer.matches_provided_phone_number)

    def test_warning(self):
        code = 'INVALID_INPUT'
        msg = 'Input invalid'

        warning = ServiceWarning({
            'code': code,
            'warning': msg,
            'input_pointer': "/first/second"
        })

        self.assertEqual(code, warning.code)
        self.assertEqual(msg, warning.warning)
        self.assertEqual('/first/second', warning.input_pointer)

    def test_score(self):
        id = 'b643d445-18b2-4b9d-bad4-c9c4366e402a'
        score = Score({
            'id': id,
            'funds_remaining': 10.01,
            'queries_remaining': 123,
            'risk_score': 0.01,
            'ip_address': {
                'risk': 99
            },
            'warnings': [{
                'code': 'INVALID_INPUT'
            }],
        })

        self.assertEqual(id, score.id)
        self.assertEqual(10.01, score.funds_remaining)
        self.assertEqual(123, score.queries_remaining)
        self.assertEqual(0.01, score.risk_score)
        self.assertEqual('INVALID_INPUT', score.warnings[0].code)
        self.assertEqual(99, score.ip_address.risk)

    def test_insights(self):
        response = self.factors_response()
        del response['subscores']
        insights = Insights(response)
        self.check_insights_data(insights, response['id'])

    def test_factors(self):
        response = self.factors_response()
        factors = Factors(response)
        self.check_insights_data(factors, response['id'])
        self.assertEqual(0.01, factors.subscores.avs_result)
        self.assertEqual(0.02, factors.subscores.billing_address)
        self.assertEqual(
            0.03, factors.subscores.billing_address_distance_to_ip_location)
        self.assertEqual(0.04, factors.subscores.browser)
        self.assertEqual(0.05, factors.subscores.chargeback)
        self.assertEqual(0.06, factors.subscores.country)
        self.assertEqual(0.07, factors.subscores.country_mismatch)
        self.assertEqual(0.08, factors.subscores.cvv_result)
        self.assertEqual(0.09, factors.subscores.email_address)
        self.assertEqual(0.10, factors.subscores.email_domain)
        self.assertEqual(0.11, factors.subscores.email_tenure)
        self.assertEqual(0.12, factors.subscores.ip_tenure)
        self.assertEqual(0.13, factors.subscores.issuer_id_number)
        self.assertEqual(0.14, factors.subscores.order_amount)
        self.assertEqual(0.15, factors.subscores.phone_number)
        self.assertEqual(
            0.16, factors.subscores.shipping_address_distance_to_ip_location)
        self.assertEqual(0.17, factors.subscores.time_of_day)

    def factors_response(self):
        return {
            'id': "b643d445-18b2-4b9d-bad4-c9c4366e402a",
            'disposition': {
                'action': 'reject'
            },
            'ip_address': {
                'country': {
                    'iso_code': 'US'
                }
            },
            'credit_card': {
                'is_prepaid': True,
                'brand': 'Visa',
                'type': 'debit'
            },
            'device': {
                'id': "b643d445-18b2-4b9d-bad4-c9c4366e402a"
            },
            'email': {
                'is_free': True
            },
            'shipping_address': {
                'is_in_ip_country': True
            },
            'billing_address': {
                'is_in_ip_country': True
            },
            'funds_remaining': 10.01,
            'queries_remaining': 123,
            'risk_score': 0.01,
            'subscores': {
                'avs_result': 0.01,
                'billing_address': 0.02,
                'billing_address_distance_to_ip_location': 0.03,
                'browser': 0.04,
                'chargeback': 0.05,
                'country': 0.06,
                'country_mismatch': 0.07,
                'cvv_result': 0.08,
                'email_address': 0.09,
                'email_domain': 0.10,
                'email_tenure': 0.11,
                'ip_tenure': 0.12,
                'issuer_id_number': 0.13,
                'order_amount': 0.14,
                'phone_number': 0.15,
                'shipping_address_distance_to_ip_location': 0.16,
                'time_of_day': 0.17,
            },
            'warnings': [{
                "code": "INVALID_INPUT"
            }]
        }

    def check_insights_data(self, insights, uuid):
        self.assertEqual('US', insights.ip_address.country.iso_code)
        self.assertEqual(False,
                         insights.ip_address.country.is_in_european_union)
        self.assertEqual(True, insights.credit_card.is_prepaid)
        self.assertEqual('Visa', insights.credit_card.brand)
        self.assertEqual('debit', insights.credit_card.type)
        self.assertEqual(uuid, insights.device.id)
        self.assertEqual('reject', insights.disposition.action)
        self.assertEqual(True, insights.email.is_free)
        self.assertEqual(True, insights.shipping_address.is_in_ip_country)
        self.assertEqual(True, insights.billing_address.is_in_ip_country)
        self.assertEqual(uuid, insights.id)
        self.assertEqual(10.01, insights.funds_remaining)
        self.assertEqual(123, insights.queries_remaining)
        self.assertEqual(0.01, insights.risk_score)
        self.assertEqual("INVALID_INPUT", insights.warnings[0].code)
        self.assertIsInstance(insights.warnings, tuple,
                              'warnings is a tuple, not a dict')
